//---------------------------------------------------------------------------

#pragma hdrstop

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#define Z_PREFIX
#include "zlib.h"
#undef Byte
#include "HPIFile.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

struct _HPIHeader
{
  int HPIMarker;          // 'HAPI'
  int SaveMarker;         // 'BANK' if saved gamed
  int DirectorySize;      // the size of the directory
  int HeaderKey;          // decrypt key
  int Start;              // file offset of directory
};

struct _HPIEntry
{
  int NameOffset;         // points to the file name
  int DirDataOffset;      // points to directory data
  char Flag;              // file flag
};

HPIFile::HPIFile()
{
  _file = 0;
}

HPIFile::~HPIFile()
{
  if (_file)
    fclose(_file);
}

bool HPIFile::ReadFile(const char *fname)
{
  char head[20],
       *buf;
  const char *p;
  int size,
      start;

  if (_file)
    fclose(_file);
  _file = fopen(fname, "rb");
  if (!_file)
    return false;
  fread(head, 20, 1, _file);
  if (_GetInt(head) != 0x49504148)
    return false;
  p = strrchr(fname, '\\');
  if (!p)
    p = strrchr(fname, '/');
  if (!p)
    p = fname;
  _filename = p + 1;
  size = _GetInt(head + 8);
  _key = _GetInt(head + 12);
  _key = ~((_key << 2) + (_key >> 6));
  start = _GetInt(head + 16);
  buf = new char[size];
  _GetDecrypted(buf + start, size - start, start);
  _entry.Entry.clear();
  _entry.Name = ".";
  _entry.Directory = true;
  _LoadEntry(buf, buf + start, _entry);
  delete []buf;

  return true;
}

bool HPIFile::LoadFile(const char *fname, char **buf, int *size)
{
  _Entry *entry;
  int i;
  const char *name;

  name = strrchr(fname, '/');
  if (!name)
    return false;
  name++;
  entry = _GetEntryPath(fname);
  if (!entry)
    return false;
  for(i = 0; i < (int)entry->Entry.size(); i++)
    if ((!entry->Entry[i].Directory) && (name == entry->Entry[i].Name))
    {
      *size = entry->Entry[i].File.FileSize;
      *buf = new char[*size];
      switch(entry->Entry[i].File.Flag)
      {
        case 0:
          _GetDecrypted(*buf, *size, entry->Entry[i].File.DataOffset);
          break;
        case 1:
          _GetCompressed(&entry->Entry[i].File, *buf, true);
          break;
        case 2:
          _GetCompressed(&entry->Entry[i].File, *buf, false);
          break;
      }
      return true;
    }

  return false;
}

#ifdef VCL
void HPIFile::ToTree(TTreeView *tree)
{
  TTreeNodes *nodes;
  TTreeNode *node;

  nodes = tree->Items;
  nodes->Clear();
  nodes->BeginUpdate();
  node = nodes->Add(nodes->GetFirstNode(), _filename.c_str());
  node->ImageIndex = 0;
  node->SelectedIndex = 0;
  _ToTree(nodes, node, _entry);
  tree->AlphaSort();
  nodes->EndUpdate();
}

void HPIFile::ToDir(TListView *list, const char *path)
{
  const char *FLAG[] = {"Stored", "LZ77", "ZLib"};
  _Entry *entry;
  int i,
      n;

  list->Items->Clear();
  n = 0;
  entry = _GetEntryPath(path);
  if (!entry)
    return;
  for(i = 0; i < (int)entry->Entry.size(); i++)
    if (!entry->Entry[i].Directory)
    {
      list->Items->Add();
      list->Items->Item[n]->ImageIndex = 3;
      list->Items->Item[n]->Caption = entry->Entry[i].Name.c_str();
      list->Items->Item[n]->SubItems->Add(entry->Entry[i].File.FileSize);
      list->Items->Item[n]->SubItems->Add(FLAG[entry->Entry[i].File.Flag]);
      n++;
    }
  list->AlphaSort();
}
#endif

void HPIFile::_LoadEntry(char *buf, char *s, _Entry &entry)
{
  _Entry temp;
  int i,
      entries;
  char *p;

  entries = _GetInt(s);
  s = buf + _GetInt(s + 4);
  for(i = 0; i < entries; i++)
  {
    temp.Entry.clear();
    temp.Name = _StrToLower(string(buf + _GetInt(s)));
    temp.Directory = s[8];
    p = buf + _GetInt(s + 4);
    if (temp.Directory)
      _LoadEntry(buf, p, temp);
    else
    {
      temp.File.DataOffset = _GetInt(p);
      temp.File.FileSize = _GetInt(p + 4);
      temp.File.Flag = p[8];
    }
    entry.Entry.push_back(temp);
    s += 9;
  }
}

#ifdef VCL
void HPIFile::_ToTree(TTreeNodes *nodes, TTreeNode *node, _Entry &entry)
{
  TTreeNode *n;
  int i;

  for(i = 0; i < (int)entry.Entry.size(); i++)
    if (entry.Entry[i].Directory)
    {
      n = nodes->AddChild(node, entry.Entry[i].Name.c_str());
      n->ImageIndex = 1;
      n->SelectedIndex = 2;
      _ToTree(nodes, n, entry.Entry[i]);
    }
}
#endif

HPIFile::_Entry *HPIFile::_GetEntryPath(const char *path)
{
  _Entry *entry;
  const char *c;
  char *buf,
       *p;
  int i,
      l;

  entry = &_entry;
  c = path;
  buf = new char[strlen(path) + 1];
  for(;;)
  {
    strcpy(buf, c);
    p = strchr(buf, '/');
    if (!p)
      break;
    *p = 0;
    c = p + 1;
    l = (int)entry->Entry.size();
    for(i = 0; i < l; i++)
      if ((buf == entry->Entry[i].Name) && (entry->Entry[i].Directory))
      {
        entry = &entry->Entry[i];
        break;
      }
    if (i == l)
    {
      delete []buf;
      return 0;
    }
  }
  delete []buf;

  return entry;
}

void HPIFile::_GetDecrypted(char *buf, int len, int pos)
{
  int i;

  fseek(_file, pos, SEEK_SET);
  fread(buf, len, 1, _file);
  if (_key != -1)
    for(i = 0; i < len; i++)
      buf[i] = ((pos + i) ^ _key) ^ ~buf[i];
}

void HPIFile::_GetCompressed(_File *file, char *buf, bool lz77)
{
  int chunks,
      i, j,
      comp,
      dec,
      *sizes;
  char *c,
       *p;

  chunks = file->FileSize / 65536;
  if (file->FileSize % 65536)
    chunks++;
  sizes = new int[chunks];
  fseek(_file, file->DataOffset, SEEK_SET);
  _GetDecrypted((char *)sizes, 4 * chunks, file->DataOffset);
  for(i = 0; i < chunks; i++)
    sizes[i] = _GetInt((unsigned char *)&sizes[i]);
  p = buf;
  for(i = 0; i < chunks; i++)
  {
    c = new char[sizes[i]];
    _GetDecrypted(c, sizes[i], ftell(_file));
    comp = _GetInt(c + 7);
    dec = _GetInt(c + 11);
    if (c[6])
      for(j = 0; j < comp; j++)
        c[j + 19] = (c[j + 19] - j) ^ j;
    if (lz77)
      _DecompLZ77(c + 19, p, comp);
    else
      z_uncompress(p, (unsigned long *)&dec, c + 19, comp);
    p += dec;
    delete []c;
  }

  delete []sizes;
}

void HPIFile::_DecompLZ77(const char *in, char *out, int len)
{
  char hist[4096];
  int tag,
      mask,
      ihp,
      ohp,
      inp,
      outp,
      count,
      i;

  ihp = 1;
  inp = 0;
  outp = 0;
  mask = 0x80;
  for(;;)
  {
    mask <<= 1;
    if (mask == 0x100)
    {
      mask = 1;
      tag = in[inp++];
    }
    if (mask & tag)
    {
      count = ((unsigned char *)in)[inp] +
              (((unsigned char *)in)[inp + 1] << 8);
      inp += 2;
      ohp = count >> 4;
      count = (count & 0xF) + 2;
      if (!ohp)
        return;
      if (count)
        for(i = 0; i < count; i++)
        {
          out[outp++] = hist[ohp];
          hist[ihp++] = hist[ohp++];
          ihp &= 0xFFF;
          ohp &= 0xFFF;
        }
    }
    else
    {
      out[outp++] = in[inp];
      hist[ihp++] = in[inp++];
      ihp &= 0xFFF;
    }
  }
}

unsigned int HPIFile::_GetInt(unsigned char *c)
{
  return c[0] + (c[1] << 8) + (c[2] << 16) + (c[3] << 24);
}

string &HPIFile::_StrToLower(string &s)
{
  int i;

  for(i = 0; i < (int)s.length(); i++)
    s[i] = tolower(s[i]);

  return s;
}

